import os
import traceback

import requests
from requests.exceptions import RequestException
from tenacity import (
    RetryError,
    retry,
    retry_if_result,
    stop_after_attempt,
    wait_exponential,
)

from ts_sdk.task.__util_adapters import (
    CommunicationFormat,
    select_versioned_value,
)
from ts_sdk.task.__util_task.exceptions import (
    ContainerStoppedException,
    TaskUpdateConflictException,
    FailedPollTaskException,
)
from ts_sdk.task.__util_decorators import return_on_failure, raise_on_failure
from ts_sdk.task.__util_log import (
    CONTAINER_ID_KEY,
    LEVEL_KEY,
    MESSAGE_KEY,
    TASK_ID_KEY,
)
from ts_sdk.task.__util_log import Log


platform_props_hash = os.environ.get("PLATFORM_PROPS_HASH")
task_group_hash = os.environ.get("TASK_GROUP_HASH")
container_id = os.environ.get("CONTAINER_ID")

log = Log({CONTAINER_ID_KEY: container_id})


def get_endpoint():
    return select_versioned_value(
        {
            CommunicationFormat.V0: os.environ.get("ORCHESTRATOR_ENDPOINT"),
            CommunicationFormat.V1: os.environ.get("KERNEL_ENDPOINT"),
        }
    )


@raise_on_failure(RetryError, FailedPollTaskException)
@retry(
    wait=wait_exponential(min=1),
    retry=retry_if_result(lambda v: v is None),
    stop=stop_after_attempt(5),
)
def poll_task():
    try:
        poll_url = get_endpoint() + "/task/poll"
        response = requests.post(
            poll_url,
            json={
                "platformPropsHash": platform_props_hash,
                "taskGroupHash": task_group_hash,
                "containerId": container_id,
            },
            verify=False,
            timeout=300,
        )
        if response.status_code >= 400:
            message = f"Got {response.status_code} for {poll_url}"
            log.log({LEVEL_KEY: "error", MESSAGE_KEY: message})
            if response.status_code == 409:
                raise ContainerStoppedException(message)
            return None
        return generate_task_from_response(response.json())
    except RequestException:
        log.log({LEVEL_KEY: "error", MESSAGE_KEY: traceback.format_exc()})
        return None


def trim_result(result):
    return select_versioned_value(
        {
            CommunicationFormat.V0: {
                "status": result.get("status"),
                "result": result.get("result"),
            },
            # CommunicationFormat.V2: result,  # TODO uncomment when all deployments are off 3.6.{0,1}
        }
    )


@return_on_failure(RetryError, False)
@retry(
    wait=wait_exponential(min=1),
    retry=retry_if_result(lambda v: v is False),
    stop=stop_after_attempt(5),
)
def update_task_status(task, result):
    try:
        task_id = task.get("id")
        update_url = get_endpoint() + f"/task/{task_id}/update-status"
        response = requests.post(
            update_url,
            json=trim_result(result),
            verify=False,
            timeout=300,
        )
        if response.status_code >= 400:
            message = f"Got {response.status_code} for {update_url}"
            log.log(
                {
                    LEVEL_KEY: "error",
                    TASK_ID_KEY: task.get("id"),
                    MESSAGE_KEY: message,
                }
            )
            return False
    except RequestException:
        log.log(
            {
                LEVEL_KEY: "error",
                TASK_ID_KEY: task.get("id"),
                MESSAGE_KEY: traceback.format_exc(),
            }
        )
        return False
    return True


def generate_task_from_response(body):
    if body:
        data = body.get("data")
        return {
            "id": body.get("id"),
            "context": data.get("context", {}) or {},
            "input": data.get("input", {}) or {},
            "secrets": data.get("secrets", {}) or {},
            "func": data.get("func"),
            "workflow_id": data.get("workflowId"),
            "correlation_id": body.get("correlationId") or body.get("id"),
            "func_dir": data.get("funcDir", "./func") or "./func",
        }

    return {}


def extend_task_timeout(task):
    try:
        task_id = task.get("id")
        extend_timeout_url = get_endpoint() + f"/task/{task_id}/extend-timeout"

        response = requests.post(
            extend_timeout_url,
            json={},
            verify=False,
            timeout=300,
        )
        if response.status_code >= 400:
            message = f"Got {response.status_code} for {extend_timeout_url}"
            log.log(
                {
                    LEVEL_KEY: "error",
                    TASK_ID_KEY: task.get("id"),
                    MESSAGE_KEY: message,
                }
            )
            if response.status_code == 409:
                raise TaskUpdateConflictException(message)
        log.log(
            {
                LEVEL_KEY: "debug",
                TASK_ID_KEY: task.get("id"),
                MESSAGE_KEY: f"EXTENDING: {task_id}",
            }
        )
    except RequestException:
        log.log(
            {
                LEVEL_KEY: "error",
                TASK_ID_KEY: task.get("id"),
                MESSAGE_KEY: traceback.format_exc(),
            }
        )



import boto3
import simplejson as json


class Storage:
    def __init__(self, bucket, file_key, endpoint):
        self.bucket = bucket
        self.file_key = file_key
        if endpoint:
            self.s3 = boto3.client(
                "s3",
                endpoint_url=endpoint,
                aws_access_key_id="123",
                aws_secret_access_key="abc",
            )
        else:
            self.s3 = boto3.client("s3")

    def writeObject(self, obj):
        full_file_key = self.file_key.format(**obj)
        response = self.s3.put_object(
            Bucket=self.bucket, Key=full_file_key, Body=json.dumps(obj)
        )
        print(
            {
                "level": "debug",
                "message": "storage object created",
                "response": response,
            }
        )



import random
import time
from datetime import datetime, timedelta, timezone

import requests
import simplejson as json

from ts_sdk.task.__util_adapters import CommunicationFormat, select_versioned_value
from ts_sdk.task.__util_adapters.endpoint_adapter import get_public_endpoint


def get_command_url():
    return get_public_endpoint("COMMAND_ENDPOINT") + select_versioned_value(
        {
            CommunicationFormat.V0: "/internal",
            CommunicationFormat.V1: "/v1/commands",
        }
    )


def run_command(
    context_data,
    org_slug,
    target_id,
    action,
    metadata,
    payload,
    ttl_sec,
    initial_delay_sec=0,
    return_command=False,
):
    if org_slug is None:
        raise Exception("Param org_slug is missing")
    if target_id is None:
        raise Exception("Param target_id is missing")
    if action is None:
        raise Exception("Param action is missing")
    if payload is None:
        raise Exception("Param payload is missing")
    if ttl_sec < 60 or ttl_sec > 900:
        raise Exception("Param ttl_sec must be between 60 and 900 seconds")
    if initial_delay_sec < 0 or initial_delay_sec > 60:
        raise Exception(
            "Param initial_delay_sec must be non-negative and less than 60 seconds"
        )

    if metadata is None:
        metadata = {}

    metadata["workflowId"] = context_data.get("workflowId")
    metadata["pipelineId"] = context_data.get("pipelineId")
    metadata["taskId"] = context_data.get("taskId")

    url = get_command_url()
    expires_at = datetime.now(timezone.utc) + timedelta(seconds=ttl_sec)

    command_create_payload = {
        "targetId": target_id,
        "action": action,
        "metadata": metadata,
        "expiresAt": expires_at.isoformat(),
        "payload": payload,
    }

    headers = {"x-org-slug": org_slug, "Content-Type": "application/json"}

    response = requests.post(
        url,
        headers=headers,
        data=json.dumps(command_create_payload),
        verify=False,
    )
    if response.status_code == 200:
        print("Command successfully created")
        r = json.loads(response.text)
        command_id = r.get("id")

        command_url = get_command_url() + "/" + command_id
        command_headers = {"x-org-slug": org_slug}

        time_elapsed = 0
        max_sleep_time = (
            32  # Maximum sleep time in seconds to prevent excessive waiting
        )
        jitter_range = (
            0.8,
            1.2,
        )  # Jitter range for exponential backoff, arbitrarily chosen without a specific reason
        sleep_time = 1 * random.uniform(*jitter_range)  # Initial sleep time

        # Wait for the initial delay before starting to poll
        if initial_delay_sec > 0:
            print(f"Waiting for {initial_delay_sec} seconds before polling...")
            time.sleep(initial_delay_sec)

        # We use 'time_elapsed < ttl_sec' instead of 'time_elapsed <= ttl_sec'
        # because if time_elapsed == ttl_sec, the remaining sleep time would be 0.
        # In that case, min(sleep_time, ttl_sec - time_elapsed) would result in 0,
        # causing an endless loop since no more time is added to time_elapsed.
        # By stopping when time_elapsed is equal to ttl_sec, we ensure the loop exits
        # once the total elapsed time meets or exceeds the ttl_sec.
        while time_elapsed < ttl_sec:
            # Polling with exponential backoff
            sleep_time = min(
                sleep_time, ttl_sec - time_elapsed
            )  # Ensure we don't sleep past the TTL
            print(f"Polling for command status (sleeping {sleep_time} seconds)...")
            time.sleep(sleep_time)
            time_elapsed += sleep_time

            command_response = requests.get(
                command_url, headers=command_headers, verify=False
            )
            if command_response.status_code == 200:
                command = json.loads(command_response.text)
                command_status = command.get("status")
                print("Current command status: " + command_status)
                if command_status == "SUCCESS":
                    if return_command:
                        return command
                    return command.get("responseBody")
                elif command_status in ["CREATED", "PENDING", "PROCESSING"]:
                    # Continue polling with backoff and jitter
                    jitter = random.uniform(*jitter_range)
                    sleep_time = min(
                        max_sleep_time, sleep_time * 2 * jitter
                    )  # Double sleep time, up to max_sleep_time
                    continue
                else:
                    if return_command:
                        return command
                    raise Exception(command.get("responseBody"))

        print("TTL for command has expired")
        if return_command:
            return command
        raise Exception("Command TTL has expired")
    else:
        print("Error creating command: " + response.text)
        raise Exception(response.text)



import io
import typing as t
from dataclasses import dataclass
from typing import IO

import typing_extensions as te

FileCategory = te.Literal["IDS", "RAW", "PROCESSED", "TMP"]


class File(te.TypedDict, total=False):
    type: te.Literal["s3file"]
    bucket: str
    fileKey: str
    fileId: str
    version: t.Optional[str]


class ReadResult(te.TypedDict):
    metadata: t.Mapping[str, str]
    body: t.Optional[bytes]
    file_obj: t.Optional[io.BufferedIOBase]
    download: t.Optional[str]
    custom_metadata: t.Mapping[str, str]
    custom_tags: t.List[str]


ContextType = t.Union[IO, str, t.Dict[str, t.Any]]


class IdsUtilDict(te.TypedDict):
    """A dictionary containing IDS utility functions as created in __util_ids.py"""

    get_ids: t.Callable[[str, str, str], t.Dict[str, t.Any]]
    # fmt: off
    validate_ids: t.Callable[[ContextType, t.Optional[str], t.Optional[str], t.Optional[str]], None]
    # fmt: on


class LabelDict(te.TypedDict, total=False):
    """
    The dictionary representation of a "label" on a file in TDP.
    It represents a `key: value` pair where key is stored in `Label.name` and value is stored in `Label.value`

    `Label.name` has a maximum limit of 128 characters
    `Label.value` has maximum limit of 256 characters

    For further info, see
    https://developers.tetrascience.com/docs/basic-concepts-metadata-tags-and-labels
    """

    name: str
    value: str


@dataclass
class Label:
    """The class representation of a "label" on a file in TDP.

    It represents a `key: value` pair where key is stored in `Label.name` and value is stored in `Label.value`

    `Label.name` has a maximum limit of 128 characters
    `Label.value` has maximum limit of 256 characters

    For further info, see
    https://developers.tetrascience.com/docs/basic-concepts-metadata-tags-and-labels
    """

    name: str
    value: str


AnyLabel = t.Union[Label, LabelDict]



import json
import os
import sys

from main import convert, decorate


def test_basic(shared_datadir):
    with (shared_datadir / "input.json").open() as input_file:
        result = convert(input_file)
    with (shared_datadir / "expected.json").open() as expected_file:
        expected = json.load(expected_file)
    assert result == expected


def test_decorate():
    assert decorate({"a": "1"}, "label") == {"a": "1", "label": "label"}
